// Copyright Andy Singleton, 1993
// This code is released for non-commercial use only
// For questions or upgrades contact:
// Andy Singleton, Creation Mechanics Inc.
// PO Box 248, Peterborough, NH 03458
// Internet: p00396@psilink.com
// Phone: (603) 563-7757

// *********************** CFitness Object methods **********************
// Utility class for doing GA selection
// Applies proportional (roulette), rank, or elitist selection

#include "pch.h"
#include "fitness.h"
#define new DEBUG_NEW

inline float posfitness(float v){	// return positive fitness from +- value
//    return v>0? 1+v:exp(v);
	return v>0? v:0;
}

void CFitness::add(float score,FHANDLE h) {					// add an individual
	pop[icount].rawscore=score;
	pop[icount].handle=h;
	rawsum+=score;
	if (icount!=0) {
		pmin = (pop[pmin].rawscore < score? pmin : icount);
		pmax = (pop[pmax].rawscore > score? pmax : icount);
	}
	icount++;
}

FHANDLE CFitness::roulette() {	// scale and do selection on scaled scores
	int i,j,omin,winner;
	float slot,wedge,slope,temp;
	if (!didscale) {
					// scale, according to method
			switch(method) {
			case EStraight:
				for (i=0;i<icount;i++) scalescore[i]=posfitness(pop[i].rawscore);
				break;
			case ERank:
					// rank first (insertion sort)
				for (i=0;i<icount;i++) {
				    omin=-1;
				    temp=pop[i].rawscore;
				    for (j=0;j<i;j++) {
					if (temp > pop[j].rawscore && omin < scalescore[j])
					    omin = (int)scalescore[j];
					}
				    for (j=0;j<i;j++) {
					if (scalescore[j]>omin) scalescore[j]=scalescore[j]+1;
				    }
				    scalescore[i]=omin+1;   // correct rank so far.  Now bump up.
				}
					// Now do scaling
				slope=(icount - 1 != 0? (maxratio-1)/(icount-1):0);
				for (i=0;i<icount;i++) scalescore[i]=1+slope*scalescore[i];
				break;
			case EMinToMax:
				slope=(vmax()-vmin()!=0?(maxratio-1)/(vmax()-vmin()):0);
				for (i=0;i<icount;i++) scalescore[i]=1+slope*(pop[i].rawscore-vmin());
				break;
			case EAveToMax:
				slope=(vmax()-average()!=0? (maxratio-1)/(vmax()-average()): 0);
				for (i=0;i<icount;i++) {
					slot=1+slope*(pop[i].rawscore-average());
					scalescore[i]= slot>0? slot:.01;
				}
				break;
			case ERankTopHalf:		// ranking, only top half included
				break;
			}

		scalesum=0;
		for (i=0;i<icount;i++) scalesum+=scalescore[i];
		didscale=TRUE;
	}
	if (method!=ETournament)
    {
		slot=drand48()*scalesum;
		wedge=scalescore[0];
		winner=0;
		for (i=1;slot>wedge && i<icount;i++) {
			winner++;
			wedge+=scalescore[i];
		}
	}
	else
	{					// ETOURNAMENT - return best of out maxratio
		winner=random(icount);
		for (i=1;i<maxratio;i++)
		{
        	omin=random(icount);
			if (pop[omin].rawscore > pop[winner].rawscore)
				winner=omin;
        }
	}    	
	return pop[winner].handle;
}
