// Copyright Andy Singleton, 1993
// This code is released for non-commercial use only
// For questions or upgrades contact:
// Andy Singleton, Creation Mechanics Inc.
// PO Box 248, Peterborough, NH 03458
// Internet: p00396@psilink.com
// Phone: (603) 563-7757

// Utility class for doing GA selection
// Applies proportional, rank, or elitist selection
//  FITNESS OBJECT - selects individuals by scaled scores
//   also provides statistics
//  To use:
//			CFitness CFitnessOb(scaling_method, ratio_of_max_to_min)
//			CFitnessOb.reset()
//			For each individual in selection window
//				CFitnessOb.add(fitness_score, individual_number)
//			Selected_individual = CFitnessOb.roulette()

#ifndef _FITNESS_H
#define _FITNESS_H

	// HM 8.14.93 These are included for the sake of wchrome.h


#define FHANDLE int			// individual handle.  Could be int or pointer

class FitPair {
public:
	float rawscore;		// Fitness Score
	FHANDLE handle;				// Individual handle
};

#define IMAX 200			// maximum number of individuals to score

enum {EStraight,ERank,EMinToMax,EAveToMax,ERankTopHalf,ETournament};
// EStraight - scaled fitness = raw score if >0, exp(raw score) if <0.  Weak on negatives
// ERank - Scaled fitness according to rank = 1+(maxratio-1)*rank/(icount-1)
// EMinToMax - Scaled fitness line from min to max = 1+((maxratio-1)/(max-min))*(raw_score-min)
// EAveToMax - Scaled fitness line from ave to max.  Cutoff negative values at .01
// ERankTopHalf - scaled fitness of only the top half
// ETournament - tournament fitness.  Return the best.

float fitness(float v); 	// return positive fitness from +- value

class CFitness {				// Class to score fitness and select individuals
	FitPair pop[IMAX];		// population so far
	float scalescore[IMAX];	// scaled scores
	int didscale;				// flag whether we scaled already
	int icount;					// count of individuals in array
	int method;					// scaling method, enumerated above
	int pmin;					// population minimum (number of individual)
	int pmax;					// population maximum (number of individual)
	float rawsum;				// population sum (raw)
	float scalesum;			// population sum (scaled)
	float maxratio;			// ratio of max to ave or min
public:
	CFitness(int m = EStraight,float maxr = 100)		// scaling method, maxratio
				{method = m;maxratio=maxr;reset();};
	void reset() {icount = 0;pmin=0;pmax=0;rawsum=0;didscale=0;};
	void add(float score,FHANDLE h);				// add an individual
	FHANDLE roulette();							// roulette wheel selection on scaled scores
	float average() {return icount>0? rawsum/icount:0;};
	float vmin() {return pop[pmin].rawscore;}; // minimum value
	FHANDLE imin()  {return pop[pmin].handle;};		// minimum individual
	float vmax() {return pop[pmax].rawscore;};
	FHANDLE imax() {return pop[pmax].handle;};
	int count() {return icount;};
};

#endif		// #ifndef _FITNESS_H
