// Copyright Andy Singleton, 1993
// This code is released for non-commercial use only
// For questions or upgrades contact:
// Andy Singleton, Creation Mechanics Inc.
// PO Box 248, Peterborough, NH 03458
// Internet: p00396@psilink.com
// Phone: (603) 563-7757


// GPQUICK
// sample GP problem file
// Compatible with problem DLL's for the GP-GIM professional GP system
// SSRProb = Simple Symbolic regression problem.
// Includes main function

#include "pch.h"
#include "fitness.h"
#include "chrome.h"

extern char scratch[100];
#ifdef FASTEVAL
	extern evalnode* IpGlobal;
	extern Chrome* ChromeGlobal;
#endif

class SSRProb : public Problem {
public:
	SSRProb();                                              // Add primitive functions
	float fitness(Chrome* chrome);  // Driving ingredient - GP fitness function
};

// static variables that make up the evaluation environment
// Build your environment here as statics
// where the evaluation functions can get to the data
retval john, paul, george, ringo, elvis;



//**************************************************************************
//////////////////////////  Function Objects

// Evaluation functions are separated from FUNCTION objects for speed reasons.
// Each evaluation function goes with a FUNCTION object.
// Use the OPDEF,EVAL,IP and CURCHROME defines
// and the code will recompile for different eval methods

				//REQUIRED function for GPQuick
				// Return a numeric constant from -128 to +127
OPDEF(ConstEval) {return 127-GETIDX;};
class ConstFunc : public Function {
public:
	ConstFunc() {strcpy(name,"NUMBER");argnum = 0;varnum=256;weight=300;
	evalfunc=ConstEval;};
	virtual char * getprint(Chrome* st) {sprintf(scratch,"%d",127-VARNUM(st->expr[st->ip]));return scratch;};
};

				// Add two arguments.  Note that argnum = 2
OPDEF(AddEval) {return EVAL + EVAL;}
class AddFunc: public Function {
public:
	AddFunc() {strcpy(name,"Add");argnum = 2;varnum=0;weight=100;
	evalfunc=AddEval;};
};

				// Subtract two arguments
OPDEF(SubEval) {return EVAL - EVAL;}
class SubFunc: public Function {
public:
	SubFunc() {strcpy(name,"SUB");argnum = 2;varnum=0;weight=100;
	evalfunc=SubEval;};
};

				// Multiply two arguments
OPDEF(MulEval) {retval rval= EVAL * EVAL; return BOUNDF(rval);}
class MulFunc: public Function {
public:
	MulFunc() {strcpy(name,"MUL");argnum = 2;varnum=0;weight=100;
	evalfunc=MulEval;};
};

				// "Protected" division
OPDEF(DivEval)
{
	retval numerator,denominator,rval;
	numerator = EVAL;
	denominator = EVAL;
	if (denominator !=0)
		rval = numerator/denominator;
	else
		rval=0;
	return BOUNDF(rval);
}
class DivFunc: public Function {
public:
	DivFunc() {strcpy(name,"DIV");argnum = 2;varnum=0;weight=100;
	evalfunc=DivEval;};
};

				// Return the sine of one argument
/*
OPDEF(SineEval) {return sin(EVAL);}
class SineFunc: public Function {
public:
	SineFunc() {strcpy(name,"SINE");argnum = 1;varnum=0;weight=100;
	evalfunc=SineEval;};
};
*/


OPDEF(IfEval)           // Conditional. Just thrown in for reference.
					// If(condition>0, dothis, otherwise dothat)
					// Conditionals, loops, subroutines come from
					// manipulating the instruction pointer
{
	retval rval;
	if (EVAL>0)
	{
		rval=EVAL;
		IP++;           // Jump the third expression
		CHROMEP->Traverse();
		IP--;           // And back up one (since EVAL increments)
	}
	else {
		IP++;
		CHROMEP->Traverse();    // Jump the second expression
		IP--;                                   // Back up for EVAL
		rval=EVAL;
	}
	return rval;
}

class IfFunc : public Function {        // if(condition,dothis,dothat)
public:
	IfFunc() {strcpy(name,"IF");argnum=3;varnum=0;weight=100;
	evalfunc=IfEval;};

};



//**************************************************************************
/////////////////// Problem specific functions and terminals

				// A terminal named "John" that returns a data value
OPDEF(JohnEval) {return john;}
class JohnFunc: public Function {
public:
	JohnFunc() {strcpy(name,"John");argnum = 0;varnum=0;weight=100;
	evalfunc=JohnEval;};
};

				// A terminal named "Paul" that returns a data value
OPDEF(PaulEval) {return paul;}
class PaulFunc: public Function {
public:
	PaulFunc() {strcpy(name,"Paul");argnum = 0;varnum=0;weight=100;
	evalfunc=PaulEval;};
};

				// A terminal named "George" that returns a data value
OPDEF(GeorgeEval) {return george;}
class GeorgeFunc: public Function {                     // A terminal named "George" that returns a data value
public:
	GeorgeFunc() {strcpy(name,"George");argnum = 0;varnum=0;weight=100;
	evalfunc=GeorgeEval;};
};

				// A terminal named "Ringo" that returns a data value
OPDEF(RingoEval) {return ringo;}
class RingoFunc: public Function {                      // A terminal named "Ringo" that returns a data value
public:
	RingoFunc() {strcpy(name,"Ringo");argnum = 0;varnum=0;weight=100;
	evalfunc=RingoEval;};
};

				// A terminal named "Elvis" that returns a data value
OPDEF(ElvisEval) {return elvis;}
class ElvisFunc: public Function {                      // A terminal named "Elvis" that returns a data value
public:
	ElvisFunc() {strcpy(name,"Elvis");argnum = 0;varnum=0;weight=100;
	evalfunc=ElvisEval;};
};



//**************************************************************************
///////////////////////// PROBLEM definition


SSRProb::SSRProb() : Problem()
{
	// stick in all of the functions
	AddF(new ConstFunc);    // REQUIRED as function 0
	AddF(new AddFunc);
	AddF(new SubFunc);
	AddF(new MulFunc);
	AddF(new DivFunc);
//      AddF(new SineFunc);

	AddF(new JohnFunc);
	AddF(new PaulFunc);
	AddF(new GeorgeFunc);
	AddF(new RingoFunc);
	AddF(new ElvisFunc);

	// If you need to load problem specific data for evaluation
	// Or load an environment (a map, a scenario)
	// Do it here
}


// This is the value we want to discover
// Change the formula to experiment with difficulty
// Put a data field here for regression against actual data
#define REGRESSIONVALUE john*(george-paul)+2.5*ringo



float SSRProb::fitness(Chrome* chrome)
// Regression type fitness function.
// Calculate a sum squared error over some number of cases

{
	int i;
	float f = 0;
	retval answer,diff;

	// Check some number of cases (20 here)
	for (i=0;i<20;i++)
	{
		// Set up the evaluation environment
		// for a specific case
		john=random(1000);
		paul=random(1000);
		george=random(1000);
		ringo=random(1000);
		elvis=random(1000);

		// Evaluate the expression
		answer=chrome->evalAll();

		// Sum up the squared error
		diff = (REGRESSIONVALUE) - answer;
		f-=diff*diff;
	}

	chrome->lastfitness=f;
	return f;
}



//**************************************************************************
///////////////////////////////  MAIN

int main(int, char **)
// Run SSRProp (the Simple Symbolic Regression Problem)
// With a population of size 2000
// Run for 80,000 evals, or until it finds the answer
// DOS version will exit on keyboard hit
// Every 5 seconds or 1000 evals, print the best answer so far

{
	Problem* prob=new SSRProb;
	ChromeParams* params = new ChromeParams();
	Pop* pop;

    cout << "\nGPQUICK Sample Problem\n";

	randomize();
	pop = new Pop(prob,params,2000);                // population of size 2000

	int done = 0;


	// Do for 80,000 iterations, or until we find the right answer
	while (pop->gencount<80000l && !done)  // Portable version - not interruptible
	//while (pop->gencount<80000l && !done && !kbhit())  // kbhit IS NOT PORTABLE (Borland specific)
	{
		// go for 5 seconds, 1000 evals, or until we get to less than 4 error
		pop->go_until(time(NULL)+5,1000,-4);

		// Display the best so far
		cout << "\n\nAfter " << pop->gencount << " generates, error = " << pop->BestFitness << "\n";
		cout.flush();
		pop->pop[pop->BestMember]->write(PRETTY_NONE,cout);

		done = (pop->BestFitness >=-4);
	}


	// Display the best
	cout << "\n\nFINAL RESULTS...";
	cout << "\nAfter " << pop->gencount << " generates, error = " << pop->BestFitness << "\n";
	cout.flush();
	pop->pop[pop->BestMember]->write(PRETTY_NONE,cout);

	delete pop;
	delete params;
	delete prob;
	return done? 1 : 0;
}
